@@ -106,7 +106,9 @@ impl From<io::Error> for Error {
 
 impl From<nix::Error> for Error {
     fn from(val: nix::Error) -> Self {
-        io::Error::from_raw_os_error(val as i32).into()
+        val.as_errno()
+            .map(|errno| io::Error::from_raw_os_error(errno as i32).into())
+            .unwrap_or_else(|| io::Error::new(io::ErrorKind::Other, val).into())
     }
 }
 
