@@ -7,10 +7,8 @@ use enumflags2::BitFlags;
 use serde::{Deserialize, Serialize};
 use serde_repr::{Deserialize_repr, Serialize_repr};
 use std::collections::HashMap;
-use zvariant::{OwnedValue, Value};
-use zvariant_derive::Type;
+use zvariant::{derive::Type, ObjectPath, OwnedObjectPath, OwnedValue, Value};
 
-use crate as zbus;
 use crate::{dbus_proxy, DBusError};
 
 /// Proxy for the `org.freedesktop.DBus.Introspectable` interface.
@@ -28,15 +26,27 @@ trait Properties {
     fn get(&self, interface_name: &str, property_name: &str) -> Result<OwnedValue>;
 
     /// Set a property value.
-    fn set(&self, interface_name: &str, property_name: &str, value: &Value) -> Result<()>;
+    fn set(&self, interface_name: &str, property_name: &str, value: &Value<'_>) -> Result<()>;
 
     /// Get all properties.
     fn get_all(&self, interface_name: &str) -> Result<HashMap<String, OwnedValue>>;
+
+    #[dbus_proxy(signal)]
+    fn properties_changed(
+        &self,
+        interface_name: &str,
+        changed_properties: HashMap<&str, Value<'_>>,
+        invalidated_properties: Vec<&str>,
+    ) -> Result<()>;
 }
 
-type ManagedObjects = HashMap<String, HashMap<String, HashMap<String, OwnedValue>>>;
+type ManagedObjects = HashMap<OwnedObjectPath, HashMap<String, HashMap<String, OwnedValue>>>;
 
 /// Proxy for the `org.freedesktop.DBus.ObjectManager` interface.
+///
+/// **NB:** Changes to properties on existing interfaces are not reported using this interface.
+/// Please use [`PropertiesProxy::connect_properties_changed`] to monitor changes to properties on
+/// objects.
 #[dbus_proxy(interface = "org.freedesktop.DBus.ObjectManager")]
 trait ObjectManager {
     /// The return value of this method is a dict whose keys are object paths. All returned object
@@ -48,6 +58,21 @@ trait ObjectManager {
     /// that combination of object path and interface. If an interface has no properties, the empty
     /// dict is returned.
     fn get_managed_objects(&self) -> Result<ManagedObjects>;
+
+    /// This signal is emitted when either a new object is added or when an existing object gains
+    /// one or more interfaces. The `interfaces_and_properties` argument contains a map with the
+    /// interfaces and properties (if any) that have been added to the given object path.
+    #[dbus_proxy(signal)]
+    fn interfaces_added(
+        &self,
+        object_path: ObjectPath<'_>,
+        interfaces_and_properties: HashMap<&str, HashMap<&str, Value<'_>>>,
+    ) -> Result<()>;
+
+    /// This signal is emitted whenever an object is removed or it loses one or more interfaces.
+    /// The `interfaces` parameters contains a list of the interfaces that were removed.
+    #[dbus_proxy(signal)]
+    fn interfaces_removed(&self, object_path: ObjectPath<'_>, interfaces: Vec<&str>) -> Result<()>;
 }
 
 /// Proxy for the `org.freedesktop.DBus.Peer` interface.
@@ -152,8 +177,26 @@ pub enum RequestNameReply {
     AlreadyOwner = 0x04,
 }
 
+/// The return code of the [`releaes_name`] method.
+///
+/// [`release_name`]: struct.DBusProxy.html#method.release_name
+#[repr(u32)]
+#[derive(Deserialize_repr, Serialize_repr, Type, Debug, PartialEq)]
+pub enum ReleaseNameReply {
+    /// The caller has released their claim on the given name. Either the caller was the primary
+    /// owner of the name, and the name is now unused or taken by somebody waiting in the queue for
+    /// the name, or the caller was waiting in the queue for the name and has now been removed from
+    /// the queue.
+    Released = 0x01,
+    /// The given name does not exist on this bus.
+    NonExistent = 0x02,
+    /// The caller was not the primary owner of this name, and was also not waiting in the queue to
+    /// own this name.
+    NotOwner = 0x03,
+}
+
 /// Proxy for the `org.freedesktop.DBus` interface.
-#[dbus_proxy]
+#[dbus_proxy(interface = "org.freedesktop.DBus")]
 trait DBus {
     /// Adds a match rule to match messages going through the message bus
     fn add_match(&self, rule: &str) -> Result<()>;
@@ -169,6 +212,7 @@ trait DBus {
     fn get_connection_selinux_security_context(&self, bus_name: &str) -> Result<Vec<u8>>;
 
     /// Returns the Unix process ID of the process connected to the server.
+    #[dbus_proxy(name = "GetConnectionUnixProcessID")]
     fn get_connection_unix_process_id(&self, bus_name: &str) -> Result<u32>;
 
     /// Returns the Unix user ID of the process connected to the server.
@@ -196,7 +240,7 @@ trait DBus {
     fn name_has_owner(&self, name: &str) -> Result<bool>;
 
     /// Ask the message bus to release the method caller's claim to the given name.
-    fn release_name(&self, name: &str) -> Result<()>;
+    fn release_name(&self, name: &str) -> Result<ReleaseNameReply>;
 
     /// Reload server configuration.
     fn reload_config(&self) -> Result<()>;
@@ -221,16 +265,16 @@ trait DBus {
     /// This signal indicates that the owner of a name has
     /// changed. It's also the signal to use to detect the appearance
     /// of new names on the bus.
-    //#[dbus_proxy(signal)]
-    //fn name_owner_changed(&self, name: &str, old_owner: &str, new_owner: &str);
+    #[dbus_proxy(signal)]
+    fn name_owner_changed(&self, name: &str, old_owner: &str, new_owner: &str);
 
     /// This signal is sent to a specific application when it loses ownership of a name.
-    //#[dbus_proxy(signal)]
-    //fn name_lost(&self, name: &str);
+    #[dbus_proxy(signal)]
+    fn name_lost(&self, name: &str);
 
     /// This signal is sent to a specific application when it gains ownership of a name.
-    //#[dbus_proxy(signal)]
-    //fn name_acquired(&self, name: &str);
+    #[dbus_proxy(signal)]
+    fn name_acquired(&self, name: &str);
 
     /// This property lists abstract “features” provided by the message bus, and can be used by
     /// clients to detect the capabilities of the message bus with which they are communicating.
@@ -420,6 +464,9 @@ pub enum Error {
     NotContainer(String),
 }
 
+/// Alias for a `Result` with the error type [`zbus::fdo::Error`].
+///
+/// [`zbus::fdo::Error`]: enum.Error.html
 pub type Result<T> = std::result::Result<T, Error>;
 
 impl From<zbus::MessageError> for Error {
@@ -433,6 +480,9 @@ impl From<zbus::MessageError> for Error {
                 Self::InconsistentMessage("incorrect endian".to_string())
             }
             zbus::MessageError::Io(e) => Self::IOError(e.to_string()),
+            zbus::MessageError::UnmatchedBodySignature => {
+                Self::InvalidArgs("incorrect body signature".to_string())
+            }
             zbus::MessageError::NoBodySignature => {
                 Self::InvalidSignature("missing body signature".to_string())
             }
@@ -440,15 +490,20 @@ impl From<zbus::MessageError> for Error {
                 Self::InconsistentMessage("invalid message field".to_string())
             }
             zbus::MessageError::Variant(e) => Self::InconsistentMessage(e.to_string()),
+            zbus::MessageError::MissingField => {
+                Self::InconsistentMessage("Required message field missing".to_string())
+            }
         }
     }
 }
 
 #[cfg(test)]
 mod tests {
-    use crate::fdo;
-    use crate::{Error, Message};
-    use std::convert::TryInto;
+    use crate::{fdo, Error, Message};
+    use std::{
+        convert::TryInto,
+        sync::{Arc, Mutex},
+    };
 
     #[test]
     fn error_from_zerror() {
@@ -464,4 +519,66 @@ mod tests {
         let e: fdo::Error = e.try_into().unwrap();
         assert_eq!(e, fdo::Error::TimedOut("so long".to_string()));
     }
+
+    #[test]
+    fn signal() {
+        // Register a well-known name with the session bus and ensure we get the appropriate
+        // signals called for that.
+        let conn = crate::Connection::new_session().unwrap();
+        let owner_change_signaled = Arc::new(Mutex::new(false));
+        let name_acquired_signaled = Arc::new(Mutex::new(false));
+
+        let proxy = fdo::DBusProxy::new(&conn).unwrap();
+
+        let well_known = "org.freedesktop.zbus.FdoSignalTest";
+        let unique_name = conn.unique_name().unwrap().to_string();
+        {
+            let well_known = well_known.clone();
+            let signaled = owner_change_signaled.clone();
+            proxy
+                .connect_name_owner_changed(move |name, _, new_owner| {
+                    if name != well_known {
+                        // Meant for the other testcase then
+                        return Ok(());
+                    }
+                    assert_eq!(new_owner, unique_name);
+                    *signaled.lock().unwrap() = true;
+
+                    Ok(())
+                })
+                .unwrap();
+        }
+        {
+            let signaled = name_acquired_signaled.clone();
+            // `NameAcquired` is emitted twice, first when the unique name is assigned on
+            // connection and secondly after we ask for a specific name.
+            proxy
+                .connect_name_acquired(move |name| {
+                    if name == well_known {
+                        *signaled.lock().unwrap() = true;
+                    }
+
+                    Ok(())
+                })
+                .unwrap();
+        }
+
+        proxy
+            .request_name(&well_known, fdo::RequestNameFlags::ReplaceExisting.into())
+            .unwrap();
+
+        loop {
+            proxy.next_signal().unwrap();
+
+            if *owner_change_signaled.lock().unwrap() && *name_acquired_signaled.lock().unwrap() {
+                break;
+            }
+        }
+
+        let result = proxy.release_name(&well_known).unwrap();
+        assert_eq!(result, fdo::ReleaseNameReply::Released);
+
+        let result = proxy.release_name(&well_known).unwrap();
+        assert_eq!(result, fdo::ReleaseNameReply::NonExistent);
+    }
 }
